@@ -1,5 +1,5 @@
-import pygraphviz as pgv
-import roscpp
+import genpy
+import pydot
 import rosgraph
 import roslib
 import rosnode
@@ -8,15 +8,27 @@ import rospy
 import rosservice
 import rostopic
 import socket
-import sys
 from datetime import datetime
 from flask import Markup, url_for
 from io import BytesIO
-from .util import get_object_size, str_to_msg
+from .util import get_object_size, str_to_msg, initialize_msg
+
+
+def graph_to_svg(graph):
+    img_str = graph.create_svg(prog='dot')
+    sio = BytesIO()
+    sio.write(img_str)
+    sio.seek(0)
+    svg = sio.getvalue().decode('utf-8')
+    #svg = svg.replace('xlink:', '')
+    svg = Markup(svg.replace('xlink:', '').replace('w3&#45;', 'w3-').replace('hover&#45;', 'hover-'))
+    return svg
+
 
 class Topic():
     def __init__(self, name):
         super().__init__()
+        rospy.loginfo(f"Instantiating topic with name {name}")
         self.__name = name
         self.__msg = None
         self.__msg_info = None
@@ -28,6 +40,7 @@ class Topic():
             self.type = None
         try:
             self.__data_class = roslib.message.get_message_class(topic_type_info[0])
+            self.__msg_template = initialize_msg(self.__data_class())
             self.__ros_subscriber = rospy.Subscriber(name, self.__data_class, self.__onMessage)
             self.__ros_publisher = rospy.Publisher(name, self.__data_class, queue_size=1)
         except Exception as e:
@@ -82,6 +95,12 @@ class Topic():
         self.__subs.clear()
         self.__times['state'] = datetime.now()
 
+    def unregister(self):
+        if self.__ros_subscriber is not None:
+            self.__ros_subscriber.unregister()
+        if self.__ros_publisher is not None:
+            self.__ros_publisher.unregister()
+
     def addPublisher(self, node):
         self.__pubs[node.name] = node
         self.__times['state'] = datetime.now()
@@ -104,28 +123,28 @@ class Topic():
             return self.__graph
 
         # Remember time
-        print(f'Generating graph for topic {self.__name}')
+        rospy.loginfo(f'Generating graph for topic {self.__name}')
         self.__times['graph'] = datetime.now()
 
         # Generate new graph
-        graph = pgv.AGraph(directed=True, forcelabels=True)
+        graph = pydot.Dot(graph_type='digraph', forcelabels=True)
         topic_id = 'topic_' + self.__name
         topic_label = f"{self.__name}\\n{self.__type}"
-        graph.add_node(topic_id, label=topic_label, shape='box')
+        graph.add_node(pydot.Node(topic_id, label=topic_label, shape='box'))
 
         # Add publishers
         for node_name in self.__pubs:
             node_id = 'pub_' + node_name
             node_url = url_for('node.get_node_info', name=node_name)
-            graph.add_node(node_id, label=node_name, shape='oval', URL=node_url, target='_top')
-            graph.add_edge(node_id, topic_id)
+            graph.add_node(pydot.Node(node_id, label=node_name, shape='oval', URL=node_url, target='_top'))
+            graph.add_edge(pydot.Edge(node_id, topic_id))
 
         # Add subscribers
         for node_name in self.__subs:
             node_id = 'sub_' + node_name
             node_url = url_for('node.get_node_info', name=node_name)
-            graph.add_node(node_id, label=node_name, shape='oval', URL=node_url, target='_top')
-            graph.add_edge(topic_id, node_id)
+            graph.add_node(pydot.Node(node_id, label=node_name, shape='oval', URL=node_url, target='_top'))
+            graph.add_edge(pydot.Node(topic_id, node_id))
 
         # Store and return graph
         self.__graph = graph
@@ -138,18 +157,15 @@ class Topic():
             return self.__svg
 
         # Remember time
-        print(f'Generating svg for topic {self.__name}')
+        rospy.loginfo(f'Generating svg for topic {self.__name}')
         self.__times['svg'] = datetime.now()
 
-        # Generate svg
-        img_stream = BytesIO()
-        graph.draw(path=img_stream, format='svg', prog='dot')
-        svg = img_stream.getvalue().decode('utf-8')
-        svg = svg.replace('xlink:', '')
-
         # Store and return svg
-        self.__svg = svg
-        return svg
+        self.__svg = graph_to_svg(graph)
+        return self.__svg
+
+    def msg_template(self):
+        return genpy.message.strify_message(self.__msg_template)
 
     def __onMessage(self, msg):
         if self.__msg_info is None:
@@ -172,6 +188,8 @@ class Node:
         self.__name = name
         self.__subs = dict()
         self.__pubs = dict()
+        master = rosgraph.Master(name)
+        self.__uri = master.lookupNode(name)
         self.__srvs = dict()
         self.__graph = None
         self.__svg = None
@@ -197,6 +215,10 @@ class Node:
     def services(self):
         return self.__srvs
 
+    @property
+    def uri(self):
+        return self.__uri
+
     def addSubscription(self, sub):
         self.__subs[sub.name] = sub
         self.__times['state'] = datetime.now()
@@ -215,14 +237,15 @@ class Node:
             return self.__graph
 
         # Remember time
-        print(f'Generating graph for node {self.__name}')
+        rospy.loginfo(f'Generating graph for node {self.__name}')
         self.__times['graph'] = datetime.now()
 
         # Generate new graph
-        graph = pgv.AGraph(directed=True, forcelabels=True, stylesheet='https://www.w3schools.com/w3css/4/w3.css')
+        graph = pydot.Dot(graph_type='digraph', forcelabels=True, stylesheet='https://www.w3schools.com/w3css/4/w3.css')
 
         # Add node
-        graph.add_node(self.__name, **{'shape': 'oval', 'class': 'w3-orange w3-hover-red'})
+        node_label = f"{self.__name}\\n{self.__uri}"
+        graph.add_node(pydot.Node(self.__name, label=node_label, **{'shape': 'oval', 'class': 'w3-orange w3-hover-red'}))
 
         # Add topics to which the node subscribes
         node_names = set()
@@ -230,16 +253,17 @@ class Node:
             topic_id = 'topic_sub_' + topic_name
             node_url = url_for('topic.get_topic_info', name=topic_name)
             node_label = f"{topic_name}\\n{topic.type}"
-            graph.add_node(topic_id, label=node_label, shape='box', URL=node_url, target='_top')
-            graph.add_edge(topic_id, self.__name)
+            graph.add_node(pydot.Node(topic_id, label=node_label, shape='box', URL=node_url, target='_top'))
+            graph.add_edge(pydot.Edge(topic_id, self.__name))
 
             for pub_name in topic.publishers:
                 subnode_id = 'pub_' + pub_name
                 if pub_name not in node_names:
                     subnode_url = url_for('node.get_node_info', name=pub_name)
-                    graph.add_node(subnode_id, label=pub_name, shape='oval', URL=subnode_url, target='_top')
+                    node_label = f"{pub_name}\\n{topic.publishers[pub_name].uri}"
+                    graph.add_node(pydot.Node(subnode_id, label=node_label, shape='oval', URL=subnode_url, target='_top'))
                     node_names.add(pub_name)
-                graph.add_edge(subnode_id, topic_id)
+                graph.add_edge(pydot.Edge(subnode_id, topic_id))
 
         # Add topics which the node publishes
         node_names.clear()
@@ -247,16 +271,17 @@ class Node:
             topic_id = 'topic_pub_' + topic_name
             node_url = url_for('topic.get_topic_info', name=topic_name)
             node_label = f"{topic_name}\\n{topic.type}"
-            graph.add_node(topic_id, label=node_label, shape='box', URL=node_url, target='_top')
-            graph.add_edge(self.__name, topic_id)
+            graph.add_node(pydot.Node(topic_id, label=node_label, shape='box', URL=node_url, target='_top'))
+            graph.add_edge(pydot.Edge(self.__name, topic_id))
 
             for sub_name in topic.subscribers:
                 subnode_id = 'sub_' + sub_name
                 if sub_name not in node_names:
                     subnode_url = url_for('node.get_node_info', name=sub_name)
-                    graph.add_node(subnode_id, label=sub_name, shape='oval', URL=subnode_url, target='_top')
+                    node_label = f"{sub_name}\\n{topic.subscribers[sub_name].uri}"
+                    graph.add_node(pydot.Node(subnode_id, label=node_label, shape='oval', URL=subnode_url, target='_top'))
                     node_names.add(sub_name)
-                graph.add_edge(topic_id, subnode_id)
+                graph.add_edge(pydot.Edge(topic_id, subnode_id))
 
         # Store and return graph
         self.__graph = graph
@@ -269,18 +294,12 @@ class Node:
             return self.__svg
 
         # Remember time
-        print(f'Generating svg for node {self.__name}')
+        rospy.loginfo(f'Generating svg for node {self.__name}')
         self.__times['svg'] = datetime.now()
 
-        # Generate svg
-        img_stream = BytesIO()
-        graph.draw(path=img_stream, format='svg', prog='dot')
-        svg = img_stream.getvalue().decode('utf-8')
-        svg = Markup(svg.replace('xlink:', '').replace('w3&#45;', 'w3-').replace('hover&#45;', 'hover-'))
-
         # Store and return svg
-        self.__svg = svg
-        return svg
+        self.__svg = graph_to_svg(graph)
+        return self.__svg
 
 
 class Service:
@@ -345,20 +364,20 @@ class Service:
             return self.__graph
 
         # Remember time
-        print(f'Generating graph for service {self.__name}')
+        rospy.loginfo(f'Generating graph for service {self.__name}')
         self.__times['graph'] = datetime.now()
 
         # Generate new graph
-        graph = pgv.AGraph(directed=True, forcelabels=True, stylesheet='https://www.w3schools.com/w3css/4/w3.css')
+        graph = pydot.Dot(graph_type='digraph', forcelabels=True, stylesheet='https://www.w3schools.com/w3css/4/w3.css')
         
         node_label = f"{self.__name}\\n{self.__type}"
         node_id = "srv_" + self.__name
-        graph.add_node(node_id, label=node_label, shape='box')
+        graph.add_node(pydot.Node(node_id, label=node_label, shape='box'))
 
         for _, node in self.__providers.items():
             node_url = url_for('node.get_node_info', name=node.name)
-            graph.add_node(node.name, shape='oval', URL=node_url, target='_top')
-            graph.add_edge(node.name, node_id)
+            graph.add_node(pydot.Node(node.name, shape='oval', URL=node_url, target='_top'))
+            graph.add_edge(pydot.Edge(node.name, node_id))
 
         # Store and return graph
         self.__graph = graph
@@ -371,18 +390,12 @@ class Service:
             return self.__svg
 
         # Remember time
-        print(f'Generating svg for node {self.__name}')
+        rospy.loginfo(f'Generating svg for node {self.__name}')
         self.__times['svg'] = datetime.now()
 
-        # Generate svg
-        img_stream = BytesIO()
-        graph.draw(path=img_stream, format='svg', prog='dot')
-        svg = img_stream.getvalue().decode('utf-8')
-        svg = Markup(svg.replace('xlink:', '').replace('w3&#45;', 'w3-').replace('hover&#45;', 'hover-'))
-
         # Store and return svg
-        self.__svg = svg
-        return svg
+        self.__svg = graph_to_svg(graph)
+        return self.__svg
 
 
 class Singleton(type):
@@ -395,7 +408,7 @@ class Singleton(type):
 
 class ROSApi(metaclass=Singleton):
     def __init__(self):
-        print("Instance created")
+        rospy.loginfo("Instance created")
         self.__system_state = None
         self.__nodes = dict()
         self.__topics = dict()
@@ -405,6 +418,7 @@ class ROSApi(metaclass=Singleton):
         self.__blacklisted_nodes = set()
         self.__graph = None
         self.__svg = None
+        self.__master_pid = None
         self.__times = {
             'state': datetime.now(),
             'graph': None,
@@ -437,29 +451,29 @@ class ROSApi(metaclass=Singleton):
             return self.__graph
 
         # Remember time
-        print(f'Generating overview graph')
+        rospy.loginfo(f'Generating overview graph')
         self.__times['graph'] = datetime.now()
 
         # Generate new graph
-        graph = pgv.AGraph(directed=True, forcelabels=True)
+        graph = pydot.Dot(graph_type='digraph', forcelabels=True)
 
         # Add one graph node for each ros node
         for node_name in ros.nodes:
             node_url = url_for('node.get_node_info', name=node_name)
-            graph.add_node(node_name, shape='oval', URL=node_url, target='_top')
+            graph.add_node(pydot.Node(node_name, shape='oval', URL=node_url, target='_top'))
             
         # Iterate over topics, add a graph node for each topic and draw edges to subscribers and publishers
         for topic_name, topic in ros.topics.items():
             topic_id = 'topic_' + topic_name
             node_url = url_for('topic.get_topic_info', name=topic_name)
             node_label = f"{topic_name}\\n{topic.type}"
-            graph.add_node(topic_id, label=node_label, shape='box', URL=node_url, target='_top')
+            graph.add_node(pydot.Node(topic_id, label=node_label, shape='box', URL=node_url, target='_top'))
 
             for node_name in topic.publishers:
-                graph.add_edge(node_name, topic_id)
+                graph.add_edge(pydot.Edge(node_name, topic_id))
 
             for node_name in topic.subscribers:
-                graph.add_edge(topic_id, node_name)
+                graph.add_edge(pydot.Edge(topic_id, node_name))
 
         # Store and return graph
         self.__graph = graph
@@ -472,18 +486,12 @@ class ROSApi(metaclass=Singleton):
             return self.__svg
 
         # Remember time
-        print(f'Generating overview svg')
+        rospy.loginfo(f'Generating overview svg')
         self.__times['svg'] = datetime.now()
 
-        # Generate svg
-        img_stream = BytesIO()
-        graph.draw(path=img_stream, format='svg', prog='dot')
-        svg = img_stream.getvalue().decode('utf-8')
-        svg = Markup(svg.replace('xlink:', '').replace('w3&#45;', 'w3-').replace('hover&#45;', 'hover-'))
-
         # Store and return svg
-        self.__svg = svg
-        return svg
+        self.__svg = graph_to_svg(graph)
+        return self.__svg
 
     def get_topic(self, name):
         if name in self.__topics:
@@ -508,26 +516,38 @@ class ROSApi(metaclass=Singleton):
 
         # go through the master system state first
         try:
+            pid = master.getPid()
             state = master.getSystemState()
         except socket.error:
+            pid = None
             raise rosnode.ROSNodeIOException("Unable to communicate with master!")
         
         try:
             param_names = rosparam.list_params('')
         except rosparam.RosParamIOException:
-            print('Could not fetch parameter names from server', file=sys.stdout)
+            rospy.loginfo('Could not fetch parameter names from server')
         self.__params = sorted(set(['/'.join(param.split('/')[:2]) for param in param_names]))
 
-        if state is None or state != self.__system_state:
+        if state is None or state != self.__system_state or pid != self.__master_pid:
             # Clear state
-            print('Resetting nodes and services')
+            rospy.loginfo('Resetting nodes and services')
             self.__times['state'] = datetime.now()
             
             # Topics are not deleted so that we don't loose the message statistics
             self.__nodes.clear()
             self.__services.clear()
-            for _, topic in self.__topics.items():
-                topic.clear()
+            if pid != self.__master_pid:
+                rospy.loginfo("Resetting topics since roscore PID has changed")
+                self.__master_pid = pid
+                topic_names = []
+                for name, topic in self.__topics.items():
+                    topic_names.append(name)
+                    topic.unregister()
+                for name in topic_names:
+                    del self.__topics[name]
+            else:    
+                for _, topic in self.__topics.items():
+                    topic.clear()
 
             # Iterate over publisher topics and create nodes and topics if necessary
             for s in state[0]:
